//! Input validation module

use std::fmt::Display;

use lazy_static::lazy_static;
use regex::Regex;
#[cfg(feature = "serialize")]
use serde::Serialize;

lazy_static! {
    /// Postcode regex.
    /// [Source](https://assets.publishing.service.gov.uk/government/uploads/system/uploads/attachment_data/file/488478/Bulk_Data_Transfer_-_additional_validation_valid_from_12_November_2015.pdf)
    pub static ref POSTCODE_REGEX: Regex = Regex::new("^([Gg][Ii][Rr] 0[Aa]{2})|((([A-Za-z][0-9]{1,2})|(([A-Za-z][A-Ha-hJ-Yj-y][0-9]{1,2})|(([AZa-z][0-9][A-Za-z])|([A-Za-z][A-Ha-hJ-Yj-y][0-9]?[A-Za-z])))) [0-9][A-Za-z]{2})$").unwrap();
}

/// Which address line are we talking about? Who knows?
#[derive(Debug, Clone, PartialEq, Eq)]
#[cfg_attr(feature = "serialize", derive(Serialize))]
pub enum AddressLine {
    One,
    Two,
    City,
}

/// Reason for a validation failing.
#[derive(Debug, Clone, PartialEq, Eq)]
#[cfg_attr(feature = "serialize", derive(Serialize))]
pub enum ValidationFailReason {
    Name,
    AddressLine(AddressLine),
    PostCode,
    Email,
    Password,
    EmptyString,
}

impl Display for ValidationFailReason {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::Name => write!(f, "Names must be longer than one character"),
            Self::AddressLine(AddressLine::One) => {
                write!(f, "Address Line One must be more than one character long.")
            },
            Self::AddressLine(AddressLine::Two) => {
                write!(f, "Address Line Two must be more than one character long.")
            },
            Self::AddressLine(AddressLine::City) => {
                write!(f, "City must be more than one character long.")
            },
            Self::PostCode => {
                write!(
                    f,
                    "Please enter a valid post code. Including the space in the middle."
                )
            },
            Self::Email => write!(f, "Please enter a valid email address."),
            Self::Password => write!(f, "Passwords must be 8 chars long and contain at least three of: Uppercase, Lowercase, Digit, Punctuation"),
            Self::EmptyString => write!(f, "Please provide a value."),
        }
    }
}

/// Validate that the string value is not empty.
pub fn validate_non_empty_string(test: &str) -> Result<(), ValidationFailReason> {
    if test.trim().len() > 0 {
        Ok(())
    } else {
        Err(ValidationFailReason::EmptyString)
    }
}

/// Check if a string is a valid name.
/// Must be more than one char long
pub fn is_valid_name(test: &str) -> bool {
    !test.is_empty()
}

/// Check if a string is a valid name.
/// Must be more than one char long
pub fn validate_name(test: &str) -> Result<(), ValidationFailReason> {
    if is_valid_name(test) {
        Ok(())
    } else {
        Err(ValidationFailReason::Name)
    }
}

/// Phone numbers must be 11 digits and start with 0
pub fn is_valid_phone_number(test: &str) -> bool {
    let mut chars = test.chars();

    test.len() == 11 && chars.next().unwrap_or('1') == '0' && chars.all(|c| c.is_ascii_digit())
}

/// Must be more that one char long
pub fn is_valid_address_line(test: &str) -> bool {
    !test.is_empty()
}

/// Validate an email address, returning a result.
pub fn validate_address_line(
    test: &str,
    address_line: AddressLine,
) -> Result<(), ValidationFailReason> {
    if is_valid_address_line(test) {
        Ok(())
    } else {
        Err(ValidationFailReason::AddressLine(address_line))
    }
}

/// Must be longer than one char and match [`struct@POSTCODE_REGEX`]
pub fn is_valid_postcode(test: &str) -> bool {
    !test.is_empty() && POSTCODE_REGEX.is_match(test)
}

/// Must be longer than one char and match [`struct@POSTCODE_REGEX`]
pub fn validate_postcode(test: &str) -> Result<(), ValidationFailReason> {
    if is_valid_postcode(test) {
        Ok(())
    } else {
        Err(ValidationFailReason::PostCode)
    }
}

/// Emails must be over 6 chars long
pub fn is_valid_email(username: &str) -> bool {
    // TODO: This should probably be less shit
    username.len() > 6
}

/// Emails must be over 6 chars long
pub fn validate_email(test: &str) -> Result<(), ValidationFailReason> {
    if is_valid_email(test) {
        Ok(())
    } else {
        Err(ValidationFailReason::Email)
    }
}

/// Passwords must be 8 chars long and contain at least three of: Uppercase, Lowercase, Digit,
/// Punctuation
pub fn is_valid_password(password: &str) -> bool {
    // TODO: This should probably be less shit
    let mut result = false;

    if password.len() > 8 {
        let mut score = 0;

        if password.chars().any(|c| c.is_ascii_lowercase()) {
            score += 1;
        }

        if password.chars().any(|c| c.is_ascii_uppercase()) {
            score += 1;
        }

        if password.chars().any(|c| c.is_ascii_digit()) {
            score += 1;
        }

        if password.chars().any(|c| c.is_ascii_punctuation()) {
            score += 1;
        }

        result = score >= 3;
    }

    result
}

/// Passwords must be 8 chars long and contain at least three of: Uppercase, Lowercase, Digit,
/// Punctuation
pub fn validate_password(test: &str) -> Result<(), ValidationFailReason> {
    if is_valid_password(test) {
        Ok(())
    } else {
        Err(ValidationFailReason::Password)
    }
}

#[cfg(test)]
mod test {
    macro_rules! simple_test {
        ($name:ident, $expected:expr, $test:expr, $fn:ident) => {
            #[test]
            fn $name() {
                assert_eq!($expected, $fn($test));
            }
        };
    }

    use super::{
        is_valid_address_line,
        is_valid_email,
        is_valid_name,
        is_valid_password,
        is_valid_phone_number,
        is_valid_postcode,
        validate_non_empty_string,
    };

    // validate_non_empty_string
    simple_test!(
        validate_non_empty_string_1,
        Ok(()),
        "Hello",
        validate_non_empty_string
    );

    simple_test!(
        validate_non_empty_string_2,
        Err(crate::ValidationFailReason::EmptyString),
        " ",
        validate_non_empty_string
    );

    simple_test!(
        validate_non_empty_string_3,
        Err(crate::ValidationFailReason::EmptyString),
        "  ",
        validate_non_empty_string
    );

    simple_test!(
        validate_non_empty_string_4,
        Err(crate::ValidationFailReason::EmptyString),
        "",
        validate_non_empty_string
    );

    // is_valid_name
    simple_test!(is_valid_name_1, true, "Michael", is_valid_name);
    simple_test!(is_valid_name_2, true, "m", is_valid_name);
    simple_test!(is_valid_name_3, false, "", is_valid_name);

    // is_valid_phone_number
    simple_test!(
        is_valid_phone_number_1,
        true,
        "07858298563",
        is_valid_phone_number
    );
    simple_test!(
        is_valid_phone_number_2,
        false,
        "jjjjjjjjjjj",
        is_valid_phone_number
    );
    simple_test!(
        is_valid_phone_number_3,
        false,
        "0785829856",
        is_valid_phone_number
    );
    simple_test!(
        is_valid_phone_number_4,
        false,
        "78582985633",
        is_valid_phone_number
    );
    simple_test!(
        is_valid_phone_number_5,
        false,
        "...........",
        is_valid_phone_number
    );
    simple_test!(
        is_valid_phone_number_6,
        false,
        "0785829856.",
        is_valid_phone_number
    );
    simple_test!(
        is_valid_phone_number_7,
        false,
        "0785829856k",
        is_valid_phone_number
    );

    // is_valid_address_line
    simple_test!(
        is_valid_address_line_1,
        true,
        "1 proctors way",
        is_valid_address_line
    );
    simple_test!(
        is_valid_address_line_2,
        true,
        "hello there",
        is_valid_address_line
    );
    simple_test!(is_valid_address_line_3, true, "m", is_valid_address_line);
    simple_test!(is_valid_address_line_4, false, "", is_valid_address_line);

    // is_valid_postcode
    simple_test!(is_valid_postcode_1, true, "dn20 9qf", is_valid_postcode);
    simple_test!(is_valid_postcode_2, true, "dn20 9qf", is_valid_postcode);
    simple_test!(is_valid_postcode_3, true, "DN20 9qf", is_valid_postcode);
    simple_test!(is_valid_postcode_4, true, "dn20 9QF", is_valid_postcode);
    simple_test!(is_valid_postcode_5, true, "w1 9qf", is_valid_postcode);
    simple_test!(is_valid_postcode_6, true, "s1w 9qf", is_valid_postcode);
    simple_test!(is_valid_postcode_7, false, "1ws 9qf", is_valid_postcode);
    simple_test!(is_valid_postcode_8, false, "1w 9qf", is_valid_postcode);

    // is_valid_email
    simple_test!(is_valid_email_1, true, "bruh@me.com", is_valid_email);
    simple_test!(is_valid_email_2, true, "7878787878", is_valid_email);
    simple_test!(is_valid_email_3, false, "bruh", is_valid_email);
    simple_test!(is_valid_email_4, false, "", is_valid_email);
    simple_test!(is_valid_email_5, false, "878", is_valid_email);
    simple_test!(is_valid_email_6, false, "BRUH", is_valid_email);

    // is_valid_password
    simple_test!(
        is_valid_password_1,
        true,
        "HelloThere123",
        is_valid_password
    );
    simple_test!(
        is_valid_password_2,
        true,
        "hellothere.12",
        is_valid_password
    );
    simple_test!(is_valid_password_3, true, "HELLOTHERE.1", is_valid_password);
    simple_test!(is_valid_password_4, true, "HelloThere.", is_valid_password);
    simple_test!(is_valid_password_5, false, "Hh1", is_valid_password);
    simple_test!(is_valid_password_6, false, "hellothere1", is_valid_password);
    simple_test!(is_valid_password_7, false, "HELLOTHERE1", is_valid_password);
    simple_test!(is_valid_password_8, false, "hellothere.", is_valid_password);
    simple_test!(is_valid_password_9, false, "HELLOTHERE.", is_valid_password);
    simple_test!(is_valid_password_10, false, "", is_valid_password);
}
